// Kevin Beason
// http://www.cs.fsu.edu/~beason/
// 4/17/02
//
// Progressive global illumination method. Light
// is gathered at each vertex from a random sampling of the hemisphere. In
// the first pass, only direct lighting from the luminaire is gathered. On
// later passes, these patches are now lit and able to pass light to gathering
// vertices that sample them. The lighting is purely additive and will
// eventually be infinity after enough passes.

#include <math.h>
#include <stdlib.h>
#include <iostream.h>
#include <fstream.h>
#include <string.h>
#include <stdio.h>

#include <Inventor/actions/SoWriteAction.h>
#include <Inventor/SoOutput.h>
#include <Inventor/Xt/SoXt.h>
#include <Inventor/Xt/viewers/SoXtExaminerViewer.h>
#include <Inventor/nodes/SoCoordinate3.h>
#include <Inventor/nodes/SoCube.h>
#include <Inventor/nodes/SoCone.h>
#include <Inventor/nodes/SoMaterial.h>
#include <Inventor/nodes/SoMaterialBinding.h>
#include <Inventor/nodes/SoSeparator.h>
#include <Inventor/nodes/SoSphere.h>
#include <Inventor/nodes/SoTransform.h>
#include <Inventor/actions/SoGLRenderAction.h>
#include <Inventor/sensors/SoTimerSensor.h>
#include <Inventor/nodes/SoComplexity.h>
#include <Inventor/nodes/SoCamera.h>
#include <Inventor/SoInput.h>
#include <Inventor/SoDB.h>
#include <Inventor/nodes/SoCylinder.h>
#include <Inventor/SbLinear.h>
#include <Inventor/nodes/SoShapeHints.h>
#include <Inventor/engines/SoCalculator.h>
#include <Inventor/nodes/SoVertexProperty.h>
#include <Inventor/nodes/SoIndexedFaceSet.h>
#include <Inventor/fields/SoSubField.h>
#include <Inventor/SbLinear.h>
#include <Inventor/nodes/SoIndexedLineSet.h>
#include <Inventor/events/SoKeyboardEvent.h>
#include <Inventor/nodes/SoEventCallback.h>

#include "MarchingCubes.cpp"
#include "intersect.cxx"
#include "EmissiveSampleSet.cpp"

int numVertices = 0;
int verticesPerPass = 1;
int raysPerVertex = 4;
double myGamma = 0.15;
SoCamera *Camera;
SoIndexedFaceSet *faceSet; // (SoMFInt32 coordIndex)
SoVertexProperty *vprop;   // (SoMFVec3f vertex, SoMFVec3f normal)
SoMaterial *faceMat;
SoSeparator *root;

const int valSize = 1;
const int maxDim = 256;
const int backDim[2] = {15, 15};
int dataHeaderSize = 0;
char dataFileName[] = "brain2";
int dims[3] = { 28, 28, 28 };
float data[28*28*28];
SoMFVec3f intens; // intensities 0 = dark 10000 = sun
SoMFVec3f intens2; // temporary intensity storage
const float sun_intens = 10000.0;

unsigned int
readVal(unsigned char *ptr, int valSize, int bigEndian=0){
  int i;
  unsigned int c = 0;

  if (bigEndian){
    for (i=valSize-1; i>=0; i--){
      c = (c << 8) | ptr[i];
    }
  }
  else {
    for (i=0; i<valSize; i++){
      c = (c << 8) | ptr[i];
    }
  }

  // tweak for stanford bunny
  if (c == 0xf830)
    c = 0;
  
  return c;
}

void
readScalarField(){

  // open file in binary
 
  FILE *dataFile = fopen(dataFileName, "rb");
  if (!dataFile){
    cerr << "Could not open file: " << dataFileName << endl;
    exit(1);
  }

  // skip header
  
  int c;
  unsigned char buffer[maxDim*maxDim*valSize];
  int maxc=0, minc=9999999;
  
  // iterate through each grid point
  for (int k = 0; k < dims[2]; k++) {

    // read in data
    cerr << "\rReading in slice " << k+1 << "/" << dims[2];
    fread(buffer, valSize, dims[0]*dims[1], dataFile);
    
    for (int j = 0; j < dims[1]; j++) {
      for (int i = 0; i < dims[0]; i++) {

	// compute data
	c = readVal(&buffer[(j*dims[0]+i)*valSize], valSize, 0);
	if (c<minc)
	  minc = c;
	else if (c>maxc)
	  maxc = c;
	
	data[k*dims[1]*dims[0] + j*dims[0]+i] = (float)(123-c);

      }
    }
  }
  cout << endl;
  cout << "Max: " << maxc << "  "
       << "Min: " << minc << endl;
  fclose(dataFile);
}

const SbVec3f min = SbVec3f(-1.0, -1.0, -1.0);
const SbVec3f maxDiff = SbVec3f( 2.0, 2.0, 2.0);
const SbVec3f max = min + maxDiff;

// Non-linear scale
const float gamma_val = 1000.0;
inline float nlscale(const float &x){
  return(x/gamma_val/(1.0+x/gamma_val));
}

// Non-linear scale
const inline SbColor nlscale(const SbVec3f x){
  return(SbColor(x[0]/gamma_val/(1.0+x[0]/gamma_val),
		 x[1]/gamma_val/(1.0+x[1]/gamma_val),
		 x[2]/gamma_val/(1.0+x[2]/gamma_val) ));
}

class LineSet {
private:
  SoCoordinate3 *coords;
  SoIndexedLineSet *index;
  SoSeparator *sep;
  SoMaterial *mat;
  SoMaterialBinding *bind;
  int numCoords, numIndexes;
  
public:
  LineSet(){
    coords = new SoCoordinate3();
    index = new SoIndexedLineSet();
    bind = new SoMaterialBinding();
    bind->value = SoMaterialBinding::PER_VERTEX_INDEXED;
    mat = new SoMaterial();
    sep = new SoSeparator();
    mat->diffuseColor.set1Value(0, SbColor(1.0, 0.0, 0.0)); // red
    mat->diffuseColor.set1Value(1, SbColor(1.0, 1.0, 1.0)); // white
    coords->point.set1Value(0, SbVec3f(0.0, 0.0, 0.0));
    coords->point.set1Value(1, SbVec3f(0.0, 0.0, 0.0));
    index->coordIndex.set1Value(0, 0);
    index->coordIndex.set1Value(1, 1);
    index->coordIndex.set1Value(2, -1);
    numCoords = 2;
    numIndexes = 3;
    sep->addChild(bind);
    sep->addChild(mat);
    sep->addChild(coords);
    sep->addChild(index);
  }

  SoSeparator *getSep(){
    return(sep);
  }
      
  void draw(const SbVec3f &a, const SbVec3f &b){
    // set coords
    coords->point.set1Value(numCoords++, a);
    coords->point.set1Value(numCoords++, b);

    // set index
    index->coordIndex.set1Value(numIndexes++, numCoords-2);
    index->coordIndex.set1Value(numIndexes++, numCoords-1);
    index->coordIndex.set1Value(numIndexes++, -1);

    // set color
    mat->diffuseColor.set1Value(numCoords-2, SbColor(0.0, 0.0, 1.0));
    mat->diffuseColor.set1Value(numCoords-1, SbColor(1.0, 1.0, 0.0));
  }

  void draw(const SbVec3f &a, const SbVec3f &b, const SbColor &c){
    // set coords
    coords->point.set1Value(numCoords++, a);
    coords->point.set1Value(numCoords++, b);

    // set index
    index->coordIndex.set1Value(numIndexes++, numCoords-2);
    index->coordIndex.set1Value(numIndexes++, numCoords-1);
    index->coordIndex.set1Value(numIndexes++, -1);

    // set color
    mat->diffuseColor.set1Value(numCoords-2, SbColor(0.0, 1.0, 1.0));
    mat->diffuseColor.set1Value(numCoords-1, c);
  }

  void draw1(const SbVec3f &a, const SbVec3f &b, const SbColor &c){
    // set coords
    coords->point.set1Value(0, a);
    coords->point.set1Value(1, b);

    // set color
    mat->diffuseColor.set1Value(1, c);
  }

};

//======================================================================
// This key press code is from Josh Grant (http://www.cs.fsu.edu/~grant)
//======================================================================
void keyboardEvents (void *data, SoEventCallback *eventCB)
{
  const SoEvent *ev = eventCB->getEvent();

  if (SO_KEY_PRESS_EVENT(ev, S))
    cerr << "You pressed the S key.\n" << flush;
}

LineSet *lineSet;

SoSeparator *
makeScene (){

  SoSeparator *scene = new SoSeparator();


  // This keypress code is from Josh Grant
  SoEventCallback *cb = new SoEventCallback();
  cb->addEventCallback(SoKeyboardEvent::getClassTypeId(),
                       keyboardEvents, NULL);
  scene->addChild(cb);

  
  // Cube
  SoSeparator *cubeSep = new SoSeparator();
  SoMaterial *cubeMat = new SoMaterial();
  cubeMat->diffuseColor.setValue(0.3, 1.0, 0.3);
  cubeSep->addChild(cubeMat);
  
  /*     ____
       /|1  /|5    z
      / |  / |     |
     /__|0=min4    |__y
    2|  / |6=max  /
     | /  | /    x
    3|/___|7

  */
    
  static float vertices[8][3] =
  {  { min[0], min[1], min[2]},   // 0
     { min[0], min[1], max[2]},   // 1
     { max[0], min[1], max[2]},   // 2
     { max[0], min[1], min[2]},   // 3
     { min[0], max[1], min[2]},   // 4
     { min[0], max[1], max[2]},   // 5
     { max[0], max[1], max[2]},   // 6
     { max[0], max[1], min[2]}    // 7
  };

  static int32_t indices[40] =
  {
    0, 1, 2, 3,  -1,
    2, 6, 7, 3,  -1,
    4, 7, 6, 5,  -1,
    0, 4, 5, 1,  -1,
    1, 5, 6, 2,  -1,
    0, 3, 7, 4,  -1
  };

  // coordinates
  SoCoordinate3 *cubeCoords = new SoCoordinate3();
  cubeCoords->point.setValues(0,8, vertices);
  cubeSep->addChild(cubeCoords);

  // line set
  SoIndexedLineSet *cubeLineSet = new SoIndexedLineSet();
  cubeLineSet->coordIndex.setValues(0, 40, indices);
  cubeSep->addChild( cubeLineSet );

  // lineset class
  lineSet = new LineSet();
  scene->addChild(lineSet->getSep());

  // add to scene
  scene->addChild(cubeSep);

////////////////////////////////////////////////////////////////////////  
// Begin Josh Grant's code
////////////////////////////////////////////////////////////////////////

  // create MarchingCubes engine and connect fields
  MarchingCubes *mcubes = new MarchingCubes();
  mcubes->ref();

  // set the data
  readScalarField();
  mcubes->data.setValue(dims, data, false);

  // set isoValue, in this case this means find the surface of the sphere
  // where the radius is 1.0
  mcubes->isoValue = 70.0;

  // MarchingCubes gives the triangles with the vertices ordered clockwise
  SoShapeHints *hints = new SoShapeHints();
  hints->vertexOrdering.setValue(SoShapeHints::CLOCKWISE);
  hints->shapeType.setValue(SoShapeHints::SOLID);
  scene->addChild(hints);

  // Create a VertexProperty because they are more efficient the
  // SoCoordinate3 nodes.
  vprop = new SoVertexProperty();
  vprop->ref();
  vprop->vertex.connectFrom(&mcubes->points);
  vprop->normal.connectFrom(&mcubes->normals);

  // connect the indexes from MarchingCubes
  SoSeparator *faceSep = new SoSeparator();
  faceMat = new SoMaterial();
  SoMaterialBinding *faceMatBind = new SoMaterialBinding();
  faceMatBind->value = SoMaterialBinding::PER_VERTEX_INDEXED;
  faceMat->ambientColor.setValue(0.0, 0.0, 0.0);
  faceMat->diffuseColor.setValue(0.5, 0.5, 0.5);
  faceMat->specularColor.setValue(0.0, 0.0, 0.0);
  faceMat->emissiveColor.setValue(0.0, 0.0, 0.0);
  faceMat->shininess.setValue(0.0);
  faceMat->transparency.setValue(0.0);
  faceSet = new SoIndexedFaceSet();
  faceSet->vertexProperty = vprop;
  faceSet->coordIndex.connectFrom(&mcubes->indexes);
  faceSep->addChild(faceMatBind);
  faceSep->addChild(faceMat);
  faceSep->addChild(faceSet);
  scene->addChild(faceSep);


////////////////////////////////////////////////////////////////////////  
// End Josh Grant's code
////////////////////////////////////////////////////////////////////////

  // disconnect so i can relax
  vprop->vertex.disconnect();
  vprop->normal.disconnect();
  faceSet->coordIndex.disconnect();


  // add back face polygons
  int i, j;
  int vi, vi_orig, ci;

  // create coordinates and normals
  vi = vprop->vertex.getNum();
  vi_orig = vi;
  for (j=0; j<=backDim[1]; j++)
    for (i=0; i<=backDim[0]; i++){
      vprop->vertex.set1Value(vi, SbVec3f(min[0]+maxDiff[0]*i/
					  (float)backDim[0],
					  max[1],
					  min[2]+maxDiff[2]*j/
					  (float)backDim[1]));
      vprop->normal.set1Value(vi, SbVec3f(0.0, -1.0, 0.0));
      vi++;
    }

  // create faces
  ci = faceSet->coordIndex.getNum();
  vi = vi_orig;
  for (j=0; j<backDim[1]; j++){
    for (i=0; i<backDim[0]; i++){
      faceSet->coordIndex.set1Value(ci+0, vi);
      faceSet->coordIndex.set1Value(ci+1, vi+backDim[0]+1);
      faceSet->coordIndex.set1Value(ci+2, vi+1);
      faceSet->coordIndex.set1Value(ci+3, -1);
      faceSet->coordIndex.set1Value(ci+4, vi+1);
      faceSet->coordIndex.set1Value(ci+5, vi+backDim[0]+1);
      faceSet->coordIndex.set1Value(ci+6, vi+backDim[0]+2);
      faceSet->coordIndex.set1Value(ci+7, -1);
      ci+=8;
      vi++;
    }
    vi++;
  }

  // add bottom face polygons

  // create coordinates and normals
  vi = vprop->vertex.getNum();
  vi_orig = vi;
  for (j=0; j<=backDim[1]; j++)
    for (i=0; i<=backDim[0]; i++){
      vprop->vertex.set1Value(vi, SbVec3f(min[0]+maxDiff[0]*i/
					  (float)backDim[0],
					  min[1]+maxDiff[1]*j/
					  (float)backDim[1],
					  min[2]));
      vprop->normal.set1Value(vi, SbVec3f(0.0, 0.0, 1.0));
      vi++;
    }

  // create faces
  ci = faceSet->coordIndex.getNum();
  vi = vi_orig;
  for (j=0; j<backDim[1]; j++){
    for (i=0; i<backDim[0]; i++){
      faceSet->coordIndex.set1Value(ci+0, vi);
      faceSet->coordIndex.set1Value(ci+1, vi+backDim[0]+1);
      faceSet->coordIndex.set1Value(ci+2, vi+1);
      faceSet->coordIndex.set1Value(ci+3, -1);
      faceSet->coordIndex.set1Value(ci+4, vi+1);
      faceSet->coordIndex.set1Value(ci+5, vi+backDim[0]+1);
      faceSet->coordIndex.set1Value(ci+6, vi+backDim[0]+2);
      faceSet->coordIndex.set1Value(ci+7, -1);
      ci+=8;
      vi++;
    }
    vi++;
  }

  // create light
  // set up center, center offset, normal
  const float sqd = 0.5;
  SbVec3f c = SbVec3f(0.0, -2.0, 2.0);
  SbVec3f coff = SbVec3f(-sqd*sqrt(2.0), -sqd, -sqd);
  //  SbVec3f n = -coff.cross(SbVec3f(coff[0], coff[1], -coff[2]));
  SbVec3f n = -c;
  n.normalize();
			 
  // create coordinates
  vi = vprop->vertex.getNum();
  ci = faceSet->coordIndex.getNum();
  vi_orig = vi;
  
  vprop->vertex.set1Value(vi++, c+coff);
  vprop->vertex.set1Value(vi++, SbVec3f(c[0]+coff[0],
					c[1]-coff[1],
					c[2]-coff[2]));
  vprop->vertex.set1Value(vi++, c-coff);
  vprop->vertex.set1Value(vi++, SbVec3f(c[0]-coff[0],
					c[1]+coff[1],
					c[2]+coff[2]));
  
  // set normals
  vi = vi_orig;
  vprop->normal.set1Value(vi++, n);
  vprop->normal.set1Value(vi++, n);
  vprop->normal.set1Value(vi++, n);
  vprop->normal.set1Value(vi++, n);

  // create faces
  faceSet->coordIndex.set1Value(ci++, vi_orig+0);
  faceSet->coordIndex.set1Value(ci++, vi_orig+3);
  faceSet->coordIndex.set1Value(ci++, vi_orig+1);
  faceSet->coordIndex.set1Value(ci++, -1);
  faceSet->coordIndex.set1Value(ci++, vi_orig+1);
  faceSet->coordIndex.set1Value(ci++, vi_orig+3);
  faceSet->coordIndex.set1Value(ci++, vi_orig+2);
  faceSet->coordIndex.set1Value(ci++, -1);
  

  // set intensities
  for (i = 0; i < vprop->vertex.getNum(); i++)
    intens.set1Value(i, SbVec3f(0.0, 0.0, 0.0));
  // turn light on
  for (i = vi_orig; i < vi_orig+4; i++)
    intens.set1Value(i, SbVec3f(sun_intens/1.0,
				sun_intens/1.0,
				sun_intens/1.0));
  // set intensities2
  for (i = 0; i < vprop->vertex.getNum(); i++)
    intens2.set1Value(i, intens[i]);

  // set colors
  for (i = 0; i < vprop->vertex.getNum(); i++)
    faceMat->diffuseColor.set1Value(i, SbVec3f(nlscale(intens[i][0]),
					       nlscale(intens[i][1]),
					       nlscale(intens[i][2]) ));

  // set number of vertices
  numVertices = vprop->vertex.getNum();

  return scene;
}

void debug_showray(int currVertex, int hit_idx, SbVec3f dpn){
	  
  faceMat->diffuseColor.set1Value(faceSet->coordIndex[hit_idx],
				  SbColor(0.5, 0.0, 0.0));
  faceMat->diffuseColor.set1Value(faceSet->coordIndex[hit_idx+1],
				  SbColor(0.0, 0.5, 0.0));
  faceMat->diffuseColor.set1Value(faceSet->coordIndex[hit_idx+2],
				  SbColor(0.0, 0.0, 0.5));
  faceMat->diffuseColor.set1Value(currVertex,
				  faceMat->diffuseColor[currVertex] +
				  SbColor(0.5, 0.5, 0.5));

  SoMaterial *myMat = new SoMaterial();
  myMat->diffuseColor.setValue(SbColor(1.0, 1.0, 1.0));
  SoTransform *myXform = new SoTransform();
  SoTransform *myXform2 = new SoTransform();
  myXform->translation = vprop->vertex[currVertex];
  myXform2->translation=0.1*dpn;
  SoSphere *myS = new SoSphere();
  SoSphere *myS2 = new SoSphere();
  myS->radius.setValue(0.01);
  myS2->radius.setValue(0.01);
  root->addChild(myMat);
  root->addChild(myXform);
  root->addChild(myS);
  root->addChild(myXform2);
  root->addChild(myS2);

}

void
animate(void *, SoSensor *)
{
  static int update = 1;
  static int currVertex = 0;
  static int prevVertex = 0;
  static int currPass = 0;
    int vertex = 0;
  int i;
  SbVec3f p0, n, dp, dpn, intersection, minIs;
  SbRotation sbr;
  SbVec3f receivedRadiance;
  float dot1, dot2;
  char filename[50];
  SoOutput *output;
  SoWriteAction *wa;
  SbVec3f edge1, edge2, tvec, pvec, qvec;
  double orig[3], dir[3], vert0[3], vert1[3], vert2[3];
  double t, u, v, last_t, hit_t;
  int idx, hit_idx, hit_vtx, numIndexes;
      
  if (!update)
    return;

  // loop through vertices
  for (vertex=0; vertex<verticesPerPass; vertex++){

    n = vprop->normal[currVertex];
    sbr.setValue(SbVec3f(0.0, 0.0, 1.0), n);
    receivedRadiance = SbVec3f(0.0, 0.0, 0.0);
    
    // shoot a bunch of rays
    for (i=0; i<raysPerVertex; i++){
      
      // shoot a ray
      dp.setValue(1.0, 1.0, 1.0);
      while (dp.length()>1.0){
	dp = SbVec3f(drand48()*2.0-1.0,
		     drand48()*2.0-1.0,
		     drand48());
      }
      sbr.multVec(dp, dpn);
      dpn.normalize();

      //////////////////////////////////////////////////////////////
      // Begin Triangle intersection test
      //////////////////////////////////////////////////////////////

      numIndexes = faceSet->coordIndex.getNum();
      last_t = HUGE;
      hit_idx = -1;
      
      // loop through indexes (triangles) to find closest triangle
      for (idx=0; idx<numIndexes; idx+=4){

        // copy data to data structures for intersection test
        for (int k=0; k<3; k++){
          orig[k] = vprop->vertex[currVertex][k];
          dir[k] = dpn[k];
          vert0[k] = vprop->vertex[faceSet->coordIndex[idx+0]][k];
          vert1[k] = vprop->vertex[faceSet->coordIndex[idx+1]][k];
          vert2[k] = vprop->vertex[faceSet->coordIndex[idx+2]][k];
        }

        // call intersection detection routine
        if (intersect_triangle(orig, dir, vert0, vert1, vert2, &t, &u, &v)
            && t>EPSILON && t<last_t){

	  last_t = t;
	  hit_idx = idx;
	  hit_t = t;
	}

	// end triangle search
      }

      
      //////////////////////////////////////////////////////////////
      // End Triangle intersection test
      //////////////////////////////////////////////////////////////

      // Add radiance (need to implement bilinear interpolation on
      //               struck triangle)
      if (hit_idx>=0){
	hit_vtx = faceSet->coordIndex[hit_idx];

	dot1 = vprop->normal[hit_vtx].dot(-dpn);
	if (dot1<0.0)
	  dot1 = 0.0;
	
	dot2 = vprop->normal[currVertex].dot(dpn);
	if (dot2<0.0)
	  cout << "dot2 < 0\n";
	
	receivedRadiance +=
	  (intens[hit_vtx] * dot1 * dot2)/raysPerVertex;

	if (intens[hit_vtx][0] > EPSILON ||
	    intens[hit_vtx][1] > EPSILON ||
	    intens[hit_vtx][2] > EPSILON )
	  lineSet->draw(vprop->vertex[currVertex],
			vprop->vertex[currVertex]+dpn*hit_t,
			nlscale(intens[hit_vtx]));

      }

    }

    // add radiance
    intens2.set1Value(currVertex, intens2[currVertex] + receivedRadiance*1.0);

    // draw red highlight (and fix prev highlight)
    if (vertex==verticesPerPass-1){
      // change current color
      faceMat->diffuseColor.set1Value(currVertex,
				      (nlscale(intens2[currVertex]) +
				       SbColor(1.0, 0.0, 0.0))/2.0);
      lineSet->draw1(vprop->vertex[currVertex],
		     vprop->vertex[currVertex]+dpn*((hit_idx>=0)?hit_t:10.0),
		     SbColor(1.0, 1.0, 1.0));
    
      // reset previous color
      faceMat->diffuseColor.set1Value(prevVertex,
				      nlscale(intens2[prevVertex]));
      prevVertex = currVertex;
    }
    // or just update color
    else {
      faceMat->diffuseColor.set1Value(currVertex,
				      nlscale(intens2[currVertex]));
    }

    // status update
    if (currVertex%100 == 0){
      cerr << "\rComputing pass " << currPass << " : "
	   << (int)(currVertex/(float)numVertices*100.0) << "%";
    }

    // increment vertex
    currVertex++;
    
    // if reach end, update colors
    if (currVertex >= numVertices){
      
      cerr << "\rComputing pass " << currPass << " : "
	   << "100%";
      cout << "\nPass " << currPass << " complete.\n\n";

      // swap intensity buffers
      currVertex = 0;
      for (i=0; i<numVertices; i++){
	// swap buffers
	intens.set1Value(i, intens2[i]);
	// color material
	faceMat->diffuseColor.set1Value(i, SbColor(nlscale(intens[i][0]),
						   nlscale(intens[i][1]),
						   nlscale(intens[i][2])));
      }

      // write out scene
      lineSet->draw1(vprop->vertex[currVertex],
		     vprop->vertex[currVertex],
		     SbColor(0.0, 0.0, 0.0));
      output = new SoOutput();
      sprintf(filename, "pass%d.iv", currPass+1);
      output->openFile(filename);
      wa = new SoWriteAction(output);
      wa->apply(root);
      output->closeFile();

      // increment pass 
      currPass++;
    }

    // proceed to next vertex  // check intensities
    for (i=0; i< vprop->vertex.getNum(); i++)
      if (intens[i][0] < 0.0 ||
	  intens[i][1] < 0.0 ||
	  intens[i][2] < 0.0)
	cout << "something's messed\n";

  }

  return;
}

///////////////////////////////////////////////////////////////////
// This code is from Dr. Banks - http://www.cs.fsu.edu/~banks/
//
// The main program gets a window from X, 
// puts an interactive 3D examinerViewer in it, 
// attaches the scene graph to the viewer,
// renders the scene graph,
// and hands control over to the Inventor main loop.

void
main(int argc, char* argv[])
{

  // read command line arguments
  int i;
  for (i=1; i<argc; i++){
    if (strcmp(argv[i], "-g") == 0)
      myGamma = atof(argv[i+1]);
    else if (strcmp(argv[i], "-h") == 0){
      cerr << "\nUsage: " << argv[0]
	   << " [ -h ] [ -g gamma ]\n\n";
      exit(1);
    }
  }

    
  // Initialize Inventor and Xt
  Widget myWindow = SoXt::init(argv[0]);

  // Initialize new classes
  srand48(665);
  
  if (myWindow == NULL) exit(1);

  // SoGLRenderAction *rend = new SoGLRenderAction();
  //rend->setTransparencyType(SoGLRenderAction::SORTED_OBJECT_ADD);
  
  // Create a scene graph
  root = makeScene();
   

  SoTimerSensor *sensor = new SoTimerSensor(animate, root);
  sensor->setInterval(SbTime(0.015)); // 70 frames per second
  sensor->schedule();
  
  SoXtExaminerViewer *myViewer = new SoXtExaminerViewer(myWindow);
  myViewer->setSceneGraph(root);
  Camera = myViewer->getCamera();
  myViewer->setTitle("Hi mom");
  myViewer->setTransparencyType(SoGLRenderAction::SORTED_OBJECT_BLEND);
  //  myViewer->setBackgroundColor(SbColor(0.035, 0.035, 0.035));
  myViewer->setBackgroundColor(SbColor(0.035, 0.35, 0.035));
  myViewer->show();

  // write out scene
  SoOutput *output;
  SoWriteAction *wa;
  output = new SoOutput();
  output->openFile("pass0.iv");
  wa = new SoWriteAction(output);
  wa->apply(root);
  output->closeFile();

  //SoWriteAction wa;  // Write out the scene graph just for demo purposes 
  //wa.apply(root);    // The scenegraph can be displayed using
		     // ivview, SceneViewer, or gview .
		     // It can also be displayed in a Web browser by
		     // changing its header from
		     //
		     // #Inventor V2.1 ascii
		     // 
		     // to
		     //
		     // #VRML V1.0 ascii
		     //
		     // to exploit the (deprecated) original version
		     // of the Virtual Reality Modelling Language
		     // file format.
		     // 
		     // Visit http://www.vrml.org/ (deprecated URL) or
		     // http://www.web3d.org/ for info on 
		     // VRML begat Web3D begat X3D.


  SoXt::show(myWindow);
  SoXt::mainLoop();
  }
